from enum import Enum, auto


class ExposeService(Enum):
    """
    Services for expose marimo to Internet.\n
    :code:`SSH` mode use ssh for connect to :code:`serveo.net` or :code:`localhost.run` and expose marimo.

    Note:
        For using :code:`localtunnel` you must install it.
        More about on https://github.com/localtunnel/localtunnel
    """

    ssh = auto()
    localtunnel = auto()



from subprocess import Popen, PIPE


__all__ = (
    "stop_subprocesses"
)


_dont_kill_list = ["ssh-agent"]


def stop_subprocesses(*subprocesses: Popen) -> None:
    for sub in subprocesses:
        pid = sub.pid

        temp = Popen(f"ps -e | grep {pid}", shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        p_name = str(temp.communicate()[0]).split()[4].replace(
            "\n", "").replace(
            "\\n", "").replace(
            "\\n'", "").replace(
            "'", ""
        )

        temp = Popen(["kill", f"{pid}", "STOP"], stdin=PIPE, stdout=PIPE, stderr=PIPE)
        temp.wait()

        temp = Popen(f"ps -e | grep {p_name}", shell=True, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        result = str(temp.communicate()[0])[2:][:-1]

        if result and result != "":
            for line in result.split("\\n"):
                line_split = line.split()

                if line_split and line_split[3] not in _dont_kill_list:
                    temp = Popen(
                        [f"kill", f"{line_split[0]}", "STOP"], stdin=PIPE, stdout=PIPE, stderr=PIPE
                    )
                    temp.wait()



from subprocess import Popen, PIPE

import random

import os
from pathlib import Path

from .types import Password, MarimoCMD, ExposeCMD, ExporimoSession
from .params import ExposeService

from .utils import stop_subprocesses, _dont_kill_list


__all__ = (
    "Exporimo"
)


class Exporimo:

    __index_list = list(range(1000))
    __port_range = [5500, 12500]

    __default_dir = "Notes/"
    __default_wd = str(Path(__default_dir).absolute())

    __one_setup = True

    __running_session: dict[int, ExporimoSession] = {}

    @classmethod
    def start_marimo(
            cls,
            command: str,
            file: str,
            index: int = None,
            port: int = None,
            password: str = None,
            service: ExposeService = ExposeService.ssh,
            print_url: bool = True
    ) -> str:

        cls.__setup()

        file = file if file.endswith(".py") else f"{file}.py"
        index = cls.__index_list.pop(0) if index is None else index
        port = random.randint(cls.__port_range[0], cls.__port_range[1]) if port is None else port
        password = Password() if password is None else password

        url, marimo_popen, expose_popen = cls.__start(
            command=command,
            file=file,
            port=port,
            password=password,
            service=service
        )

        if print_url:
            print(f"Your url:\n{url}")

        cls.__running_session[index] = ExporimoSession(
            marimo_popen=marimo_popen,
            expose_popen=expose_popen,
            url=url
        )

        return url

    @classmethod
    def stop_marimo(cls, index: int = None) -> None:
        if index is None and len(cls.__running_session) == 1:
            index = 0

        elif index is None:
            raise ValueError("Enter session index")

        stop_subprocesses(
            cls.__running_session[index].marimo_popen,
            cls.__running_session[index].expose_popen
        )

    @classmethod
    def wait(cls, index: int = None, until_input: bool = True, stop_word: str = "stop") -> None:
        try:
            while True:
                if until_input and input() == stop_word:
                    raise KeyboardInterrupt

        except KeyboardInterrupt:
            cls.stop_marimo(index)

        except Exception:
            cls.stop_marimo(index)
            raise

    @classmethod
    def set_port_range(cls, start: int, end: int) -> None:
        cls.__port_range[0] = start
        cls.__port_range[1] = end

    @staticmethod
    def dont_stop_serves(serves_name: str) -> None:
        _dont_kill_list.append(serves_name)

    @classmethod
    def __start(
            cls,
            *,
            command: str,
            file: str,
            port: int,
            password: str,
            service: ExposeService
    ) -> tuple[str, Popen, Popen]:

        marimo_cmd = MarimoCMD(
            command=command,
            file=file,
            port=port,
            password=password
        )
        expose_cmd, url = ExposeCMD(
            service=service,
            port=port,
            password=password
        )

        marimo_popen = Popen(marimo_cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        expose_popen = Popen(expose_cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)

        return url, marimo_popen, expose_popen

    @classmethod
    def __setup(cls) -> None:
        if cls.__one_setup:
            cls.__check_def_dir()
            cls.__check_wd()

            cls.__one_setup = False

    @classmethod
    def __check_def_dir(cls) -> None:

        path = Path(cls.__default_dir).absolute()

        if not path.exists() or (path.exists() and not path.is_dir()):
            os.mkdir(f"{path}/")

    @classmethod
    def __check_wd(cls) -> None:
        temp = Popen(
            ["pwd"],
            stdin=PIPE, stdout=PIPE, stderr=PIPE
        )
        result = str(temp.communicate()[0])[2:][:-1][:-1][:-1]

        if result != cls.__default_wd:
            os.chdir(cls.__default_wd)



import string
import random

from subprocess import Popen, PIPE, TimeoutExpired

from typing import NamedTuple

from .params import ExposeService
from .utils import stop_subprocesses


__all__ = (
    "Password",
    "MarimoCMD",
    "ExposeCMD",
    "ExporimoSession",
)


class ExporimoSession(NamedTuple):
    marimo_popen: Popen
    expose_popen: Popen
    url: str


class Password:

    # Store all characters in lists
    __letters_low: list[str] = list(string.ascii_lowercase)
    __letters_up: list[str] = list(string.ascii_uppercase)
    __digits: list[str] = list(string.digits)

    def __new__(cls, length: int = 24) -> str:

        # Shuffle all lists
        random.shuffle(cls.__letters_low)
        random.shuffle(cls.__letters_up)
        random.shuffle(cls.__digits)

        # Calculate 30% & 20% of number of characters
        first_part = round(length * (30 / 100))
        second_part = round(length * (40 / 100))

        # Generation of the password (60% letters and 40% digits)
        result: list[str] = []

        for x in range(first_part):
            result.append(cls.__letters_low[x])
            result.append(cls.__letters_up[x])

        for x in range(second_part):
            result.append(cls.__digits[x])

        # Shuffle result
        random.shuffle(result)

        # Join result
        password = "".join(result)

        return password


class MarimoCMD:

    def __new__(cls, command: str, file: str, port: int, password: str) -> list[str]:
        return ["marimo", command, file, "-p", f"{port}", "--token-password", password]


class SSHCMD:

    service_dict = {
        "serveo.net": "serveo.net",
        "localhost.run": "nokey@localhost.run"
    }

    def __init__(self):
        self.__domain = None

    def __call__ (self, port: int) -> list[str]:
        service = self.__available_service(port)
        return service

    @property
    def domain(self) -> str:
        return self.__domain

    def __available_service(self, port: int) -> list[str]:
        cmd_list = []

        for service in self.service_dict.values():
            cmd_list.append(
                ["ssh", "-R", f"{port}:localhost:{port}", service]
            )

        available_list: list[bool] = []

        for cmd in cmd_list:
            result = self.__check(cmd)
            available_list.append(result)

        if True in available_list:
            index = available_list.index(True)
            self.__domain = list(self.service_dict.keys())[index]
            return cmd_list[index]

        raise RuntimeError("All services is not available")

    @classmethod
    def __check(cls, cmd: list[str]) -> bool:
        popen = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)

        try:
            popen.communicate(timeout=0.5)
            stop_subprocesses(popen)
            return False

        except TimeoutExpired:
            stop_subprocesses(popen)
            return True


class ExposeCMD:

    def __new__(cls, service: ExposeService, port: int, password: str) -> tuple[list[str], str]:
        if service == ExposeService.ssh:
            ssh_cmd = SSHCMD()
            expose_cmd = ssh_cmd(port=port)

            host = f"{ssh_cmd.domain}:{port}"

        else:
            raise ValueError("Another services now not supported")

        return expose_cmd, cls.__url(host, password)

    @classmethod
    def __url(cls, host: str, password: str) -> str:
        return f"http://{host}?access_token={password}"



from setuptools import setup, find_packages


__version__ = "0.1.0"

with open("README.md", "r", encoding="UTF-8") as file:
    long_description = file.read()

requires_list = [
    "marimo>=0.10.17"
]

setup(
    name="exporimo",
    version=__version__,
    author="Vyacheslav Pervakov",
    author_email="WsrrcalzWehgwmD@protonmail.com",
    description="Library for fast expose marimo notebook to Internet",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FailProger/exporimo.git",
    project_urls={
        "GitHub": "https://github.com/FailProger/exporimo.git",
        "PyPI": "https://pypi.org/project/exporimo/"
    },
    license="MIT License",
    license_file="LICENSE",
    keywords=["Python", "marimo"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ],
    packages=find_packages(),
    python_requires = ">=3.13",
    install_requires=requires_list
)
